# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray as nparray
from numba import njit, prange
from typing import Union, Tuple
from collections import Iterable
__cache = True

ArrayOrFloat = Union[float, np.ndarray, list]


def itype_of_ftype(dtype):
    name = np.dtype(dtype).name
    if '32' in name:
        return np.int32
    elif '64' in name:
        return np.int64
    else:
        raise TypeError
    

@njit(nogil=True, cache=__cache)
def minmax(a: nparray) -> Tuple[float]:
    """
    Returns the minimum and maximum values of an array.
    """
    return a.min(), a.max()


def ascont(array: nparray) -> nparray:
    """
    Return a contiguous array (ndim >= 1) in memory (C order)
    """
    return np.ascontiguousarray(array)


@njit(nogil=True, cache=__cache)
def clip1d(a: nparray, a_min: float, a_max: float) -> nparray:
    """
    Clips the values outside the interval [a_min, a_max] to 
    either a_min or a_max.
    """
    a[a < a_min] = a_min
    a[a > a_max] = a_max
    return a


def atleastnd(a: nparray, n=2, front=True, back=False) -> nparray:
    """
    Returns an array that is at least 'n' dimensional.
    The required shape is obtained by inserting new axes either
    before or after the existing data. This behaviour can be controlled
    using the parameters 'front' and 'back'. If front is True and back 
    is False, new axes are crated before the first existing data index,
    and the opposite happens in every other case.
    """
    shp = a.shape
    nD = len(shp)
    if nD >= n:
        return a
    else:
        if front and not back:
            newshape = (n - nD) * (1,) + shp
        else:
            newshape = shp + (n - nD) * (1,)
        return np.reshape(a, newshape)


def atleast1d(a: ArrayOrFloat) -> nparray:
    """Returns an array that is at least 1 dimensional."""
    if not isinstance(a, Iterable):
        a = [a,]
    return np.array(a)


def atleast2d(a: nparray, **kwargs) -> nparray:
    """Returns an array that is at least 2 dimensional."""
    return atleastnd(a, 2, **kwargs)


def matrixform(f: np.ndarray) -> nparray:
    """
    Returns an array that is at least 2 dimensional.
    """
    size = len(f.shape)
    assert size <= 2, "Input array must be at most 2 dimensional."
    if size == 1:
        nV, nC = len(f), 1
        return f.reshape(nV, nC)
    return f


def atleast3d(a: nparray, **kwargs) -> nparray:
    """Returns an array that is at least 3 dimensional."""
    return atleastnd(a, 3, **kwargs)


def atleast4d(a: nparray, **kwargs) -> nparray:
    """Returns an array that is at least 4 dimensional."""
    return atleastnd(a, 4, **kwargs)


@njit(nogil=True, cache=__cache)
def flatten2dC(a: np.ndarray) -> nparray:
    I, J = a.shape
    res = np.zeros(I * J, dtype=a.dtype)
    ind = 0
    for i in range(I):
        for j in range(J):
            res[ind] = a[i, j]
            ind += 1
    return res


@njit(nogil=True, cache=__cache)
def flatten2dF(a: np.ndarray) -> nparray:
    I, J = a.shape
    res = np.zeros(I * J, dtype=a.dtype)
    ind = 0
    for j in range(J):
        for i in range(I):
            res[ind] = a[i, j]
            ind += 1
    return res


def flatten2d(a: np.ndarray, order: str = 'C') -> nparray:
    if order == 'C':
        return flatten2dC(a)
    elif order == 'F':
        return flatten2dF(a)


def isfloatarray(a: np.ndarray):
    return np.issubdtype(a.dtype, float)


def isintegerarray(a: np.ndarray):
    return np.issubdtype(a.dtype, int)


def isintarray(a: np.ndarray):
    return isintegerarray(a)


def isboolarray(a: np.ndarray):
    return np.issubdtype(a.dtype, bool)


def is1dfloatarray(a: np.ndarray):
    return isfloatarray(a) and len(a.shape) == 1


def is1dintarray(a: np.ndarray):
    return isintarray(a) and len(a.shape) == 1


def issymmetric(a: np.ndarray, tol=1e-8):
    return np.linalg.norm(a-a.T) < tol


def bool_to_float(a: np.ndarray, true=1.0, false=0.0) -> nparray:
    res = np.full(a.shape, false, dtype=float)
    res[a] = true
    return res


def choice(choices, size, probs=None) -> nparray:
    """
    Returns a numpy array, whose elements are selected from
    'choices' under probabilities provided with 'probs' (optionally).

    Example
    -------
    >>> N, p = 10, 0.2
    >>> choice([False, True], (N, N), [p, 1-p])
    """
    if probs is None:
        probs = np.full((len(choices),), 1/len(choices))
    return np.random.choice(a=choices, size=size, p=probs)


def isposdef(A: np.ndarray, tol=0):
    return np.all(np.linalg.eigvals(A) > tol)


def ispossemidef(A: np.ndarray):
    return np.all(np.linalg.eigvals(A) >= 0)


def random_pos_semidef_matrix(N) -> nparray:
    """
    Returns a random positive semidefinite matrix of shape (N, N).

    Example
    -------
    >>> random_pos_semidef_matrix(3)
    """
    A = np.random.rand(N, N)
    return A.T @ A


def random_posdef_matrix(N, alpha=1e-12) -> nparray:
    """
    Returns a random positive definite matrix of shape (N, N).
    
    All eigenvalues of this matrix are >= alpha.

    Example
    -------
    >>> random_posdef_matrix(3, 0.1)
    """
    A = np.random.rand(N, N)
    return A @ A.T + alpha*np.eye(N)


@njit(nogil=True, parallel=True, cache=__cache)
def repeat(a: np.ndarray, N=1) -> nparray:
    res = np.zeros((N, a.shape[0], a.shape[1]), dtype=a.dtype)
    for i in prange(N):
        res[i, :, :] = a
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def repeat1d(a: np.ndarray, N=1) -> nparray:
    M = a.shape[0]
    res = np.zeros(N * M, dtype=a.dtype)
    for i in prange(N):
        res[i*M : (i+1)*M] = a
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tile(a: np.ndarray, da: np.ndarray, N=1) -> nparray:
    res = np.zeros((N, a.shape[0], a.shape[1]), dtype=a.dtype)
    for i in prange(N):
        res[i, :, :] = a + i*da
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tile1d(a: np.ndarray, da: np.ndarray, N=1) -> nparray:
    M = a.shape[0]
    res = np.zeros(N * M, dtype=a.dtype)
    for i in prange(N):
        res[i*M : (i+1)*M] = a + i*da
    return res


def indices_of_equal_rows(x: np.ndarray, y: np.ndarray, tol=1e-12):
    nX, dX = x.shape
    nY, dY = y.shape
    assert dX == dY, "Input arrays must have identical second dimensions."
    square = nX == nY
    integer = isintegerarray(x) and isintegerarray(y)
    if integer:
        if square:
            inds = np.flatnonzero((x == y).all(1))
            return inds, np.copy(inds)
        else:
            return indices_of_equal_rows_njit(x, y, 0)
    else:
        if square:
            return indices_of_equal_rows_square_njit(x, y, tol)
        else:
            return indices_of_equal_rows_njit(x, y, tol)


@njit(nogil=True, parallel=True, cache=__cache)
def indices_of_equal_rows_njit(x: np.ndarray, y: np.ndarray, tol=1e-12):
    R = np.zeros((x.shape[0], y.shape[0]), dtype=x.dtype)
    for i in prange(R.shape[0]):
        for j in prange(R.shape[1]):
            R[i, j] = np.sum(np.abs(x[i] - y[j]))
    return np.where(R <= tol)


@njit(nogil=True, parallel=True, cache=__cache)
def indices_of_equal_rows_square_njit(x: np.ndarray, y: np.ndarray,
                                      tol=1e-12):
    n = np.min([x.shape[0], y.shape[0]])
    R = np.zeros((n, n), dtype=x.dtype)
    for i in prange(n):
        for j in prange(n):
            R[i, j] = np.sum(np.abs(x[i] - y[j]))
    return np.where(R <= tol)


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def count_cols(arr: np.ndarray):
    n = len(arr)
    res = np.zeros(n, dtype=np.int64)
    for i in prange(n):
        res[i] = len(arr[i])
    return res


# !FIXME : assumes a unique input array
@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def find1d(arr, space):
    res = np.zeros(arr.shape, dtype=np.uint64)
    for i in prange(arr.shape[0]):
        res[i] = np.where(arr[i] == space)[0][0]
    return res


# !TODO generalize this up to n dimensions
# !FIXME : assumes a unique input array
@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def find2d(arr, space):
    res = np.zeros(arr.shape, dtype=np.uint64)
    for i in prange(arr.shape[0]):
        res[i] = find1d(arr[i], space)
    return res


# -*- coding: utf-8 -*-
from numba import njit, prange
import numpy as np
from numpy import ndarray

__cache = True

__all__ = ['to_range']


@njit(nogil=True, parallel=True, cache=__cache)
def _to_range(vals: ndarray, source: ndarray, target: ndarray):
    res = np.zeros_like(vals)
    s0, s1 = source
    t0, t1 = target
    b = (t1- t0) / (s1 - s0)
    a = (t0 + t1) / 2 - b * (s0 + s1) / 2
    for i in prange(res.shape[0]):
        res[i] = a + b * vals[i]
    return res


def to_range(vals: ndarray, *args, source: ndarray, target: ndarray=None, 
             squeeze=False, **kwargs):
    if not isinstance(vals, ndarray):
        vals = np.array([vals,])
    source = np.array([0., 1.]) if source is None else np.array(source)
    target = np.array([-1., 1.]) if target is None else np.array(target)
    if squeeze:
        return np.squeeze(_to_range(vals, source, target))
    else:
        return _to_range(vals, source, target)


# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from sympy.physics.vector import ReferenceFrame as SymPyFrame
from typing import Iterable


from .utils import transpose_dcm_multi
from ...array import repeat
from ..array import Array


class ReferenceFrame(Array):
    """
    A base reference-frame for orthogonal vector spaces. 
    It facilitates tramsformation of tensor-like quantities across 
    different coordinate frames.

    The class is basically an interface on the `ReferenceFrame` class 
    in `sympy.physics.vector`, with a similarly working `orient_new` function.

    Examples
    --------
    Define a standard Cartesian frame and rotate it around axis 'Z'
    with an amount of 180 degrees:

    >>> A = ReferenceFrame(dim=3)
    >>> B = A.orient_new('Body', [0, 0, np.pi], 'XYZ')

    To create a third frame that rotates from B the way B rotates from A, we
    can do

    >>> A = ReferenceFrame(dim=3)
    >>> C = A.orient_new('Body', [0, 0, 2*np.pi], 'XYZ')

    or we can define it relative to B (this literally makes C to looke 
    in B like B looks in A)

    >>> C = ReferenceFrame(B.axes, parent=B)

    Notes
    -----
    The `dewloosh.geom.CartesianFrame` class takes the idea of the reference 
    frame a step further by introducing the idea of the 'origo'. 

    """

    def __init__(self, axes: ndarray = None, parent=None, *args,
                 order: str = 'row', name: str = None, dim: int = None, 
                 **kwargs):
        order = 'C' if order in ['row', 'C'] else 'F'
        try:
            if not isinstance(axes, ndarray):
                if isinstance(dim, Iterable):
                    if len(dim) == 1:
                        axes = np.eye(dim[0])
                    elif len(dim) == 2:
                        axes = repeat(np.eye(dim[-1]), dim[0])
                    else:
                        raise NotImplementedError
                elif isinstance(dim, int):
                    axes = np.eye(dim)
        except Exception as e:
            raise e
        super().__init__(axes, *args, order=order, **kwargs)
        self.name = name
        self.parent = parent
        self._order = 0 if order == 'C' else 1

    @classmethod
    def eye(cls, *args, dim=3, **kwargs):
        if len(args) > 0 and isinstance(args[0], int):
            dim = args[0]
        return cls(np.eye(dim), *args, **kwargs)

    def root(self):
        if self.parent is None:
            return self
        else:
            return self.parent.root()

    @property
    def order(self):
        return 'row' if self._order == 0 else 'col'

    @property
    def axes(self):
        """
        Returns a matrix, where each row (or column) is the component array
        of a basis vector with respect to the parent frame, or ambient
        space if there is none.
        """
        return self._array

    @axes.setter
    def axes(self, value):
        if isinstance(value, np.ndarray):
            if value.shape == self._array.shape:
                self._array = value
            else:
                raise RuntimeError("Mismatch in data dimensinons!")
        else:
            raise TypeError("Only numpy arras are supported here!")

    def show(self, target: 'ReferenceFrame' = None):
        """
        Returns the components of the current frame in a target frame.
        If the target is None, the componants are returned in the ambient frame.
        """
        return self.dcm(target=target)

    def dcm(self, *args, target: 'ReferenceFrame' = None,
            source: 'ReferenceFrame' = None, **kwargs):
        """
        Returns the direction cosine matrix (DCM) of a transformation
        from a source (S) to a target (T) frame. The current frame can be the 
        source or the target, depending on the arguments. 

        If called without arguments, it returns the DCM matrix from the 
        root frame to the current frame (S=root, T=self).

        If `source` is not `None`, then T=self.

        If `target` is not `None`, then S=self.

        Parameters
        ----------
        source : 'ReferenceFrame', Optional
            Source frame. Default is None.

        target : 'ReferenceFrame', Optional
            Target frame. Default is None.

        Returns:
        --------        
        numpy.ndarray
            DCM matrix from S to T.

        """
        if source is not None:
            # source must be a 3x3 array
            S, T = source.dcm(), self.dcm() 
            return T @ S.T
        elif target is not None:
            # target must be a 3x3 array
            S, T = self.dcm(), target.dcm()
            return T @ transpose_dcm_multi(S)
        # We only get here if the function is called without arguments.
        # The dcm from the ambient frame to the current frame is returned.
        if self.parent is None:
            return self.axes
        else:
            # parent should (but not must) be a 3x3 array
            return self.axes @ self.parent.dcm()

    def orient(self, *args, **kwargs) -> 'ReferenceFrame':
        """
        Orients the current frame inplace. 
        See `Referenceframe.orient_new` for the possible arguments.
        
        Parameters
        ----------
        args : tuple, Optional
            A tuple of arguments to pass to the `orientnew` 
            function in `sympy`. 

        kwargs : dict, Optional
            A dictionary of keyword arguments to pass to the 
            `orientnew` function in `sympy`. 

        """
        source = SymPyFrame('source')
        target = source.orientnew('target', *args, **kwargs)
        dcm = np.array(target.dcm(source).evalf()).astype(float)
        self._array = dcm @ self.axes
        return self

    def orient_new(self, *args, name='', **kwargs) -> 'ReferenceFrame':
        """
        Returns a new frame, oriented relative to the called object. 
        The orientation can be provided by all ways supported in 
        `sympy.orientnew`.

        Parameters
        ==========

        name : str
            Name for the new reference frame.

        rot_type : str
            The method used to generate the direction cosine matrix. Supported
            methods are:

            - ``'Axis'``: simple rotations about a single common axis
            - ``'DCM'``: for setting the direction cosine matrix directly
            - ``'Body'``: three successive rotations about new intermediate
              axes, also called "Euler and Tait-Bryan angles"
            - ``'Space'``: three successive rotations about the parent
              frames' unit vectors
            - ``'Quaternion'``: rotations defined by four parameters which
              result in a singularity free direction cosine matrix

        amounts :
            Expressions defining the rotation angles or direction cosine
            matrix. These must match the ``rot_type``. See examples below for
            details. The input types are:

            - ``'Axis'``: 2-tuple (expr/sym/func, Vector)
            - ``'DCM'``: Matrix, shape(3, 3)
            - ``'Body'``: 3-tuple of expressions, symbols, or functions
            - ``'Space'``: 3-tuple of expressions, symbols, or functions
            - ``'Quaternion'``: 4-tuple of expressions, symbols, or
              functions

        rot_order : str or int, optional
            If applicable, the order of the successive of rotations. The string
            ``'123'`` and integer ``123`` are equivalent, for example. Required
            for ``'Body'`` and ``'Space'``.

        Returns:
        ========      

        ReferenceFrame
            A new ReferenceFrame object.

        """
        source = SymPyFrame('source')
        target = source.orientnew('target', *args, **kwargs)
        dcm = np.array(target.dcm(source).evalf()).astype(float)
        return self.__class__(axes=dcm, parent=self, name=name)


if __name__ == '__main__':

    A = ReferenceFrame()
    B = A.orient_new('Body', [0, 0, 30*np.pi/180],  'XYZ')
    C = B.orient_new('Body', [0, 0, 30*np.pi/180],  'XYZ')



# -*- coding: utf-8 -*-
import numpy as np
from numpy.lib.mixins import NDArrayOperatorsMixin

from dewloosh.core.abc.wrap import Wrapper
from dewloosh.core.abc import ABC_Safe

from ..array import ascont

__all__ = ['ArrayBase', 'Array']


class ArrayBase(ABC_Safe, np.ndarray):

    def __new__(subtype, shape=None, dtype=float, buffer=None,
                offset=0, strides=None, order=None, frame=None,
                inds=None):
        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments.  This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to InfoArray.__array_finalize__
        obj = super().__new__(subtype, shape, dtype,
                              buffer, offset, strides, order)
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(InfoArray, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. InfoArray():
        #    obj is None
        #    (we're in the middle of the InfoArray.__new__
        #    constructor, and self.info will be set when we return 0to
        #    InfoArray.__new__)
        if obj is None:
            return
        # From view casting - e.g arr.view(InfoArray):
        #    obj is arr
        #    (type(obj) can be InfoArray)
        # From new-from-template - e.g infoarr[:3]
        #    type(obj) is InfoArray
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'info', because this
        # method sees all creation of default objects - with the
        # InfoArray.__new__ constructor, but also with
        # arr.view(InfoArray).


class Array(NDArrayOperatorsMixin, Wrapper):

    _array_cls_ = ArrayBase

    def __init__(self, *args, cls_params=None, contiguous=True, **kwargs):
        if len(args) > 0 and isinstance(args[0], np.ndarray):
            buf = ascont(args[0]) if contiguous else args[0]
        else:
            buf = np.array(*args, **kwargs)
        cls_params = dict() if cls_params is None else cls_params
        self._array = self._array_cls_(shape=buf.shape, buffer=buf,
                                       dtype=buf.dtype, **cls_params)
        super(Array, self).__init__(wrap=self._array)

    @property
    def dim(self):
        return len(self._array.shape)

    def __repr__(self):
        return f"{self.__class__.__name__}\n({self._array})"

    def __array__(self, dtype=None):
        if dtype is not None:
            return self._array.astype(dtype)
        return self._array

    def __getitem__(self, key):
        return self._array.__getitem__(key)

    def __setitem__(self, key):
        return self._array.__setitem__(key)

    def __len__(self):
        return self._array.shape[0]

    def to_numpy(self):
        return self.__array__()



# -*- coding: utf-8 -*-
import sympy as sy
import numpy as np

from .tensor import Tensor
from ._tensop import tr_3333, tr_3333_jit


class Tensor3333(Tensor):

    __imap__ = {0: (0, 0), 1: (1, 1), 2: (2, 2),
                3: (1, 2), 4: (0, 2), 5: (0, 1)}

    def __init__(self, *args, symbolic=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.symmetric = True

        imap = kwargs.get('imap', None)
        if isinstance(imap, dict):
            self.__imap__ = imap

        if 'sympy' in args or symbolic:
            self._transform = self._transform_sym
            self.dtype = object

        self.collapsed = None
        if self._array is not None:
            self.collapsed = len(self._array.shape) == 2

    def expand(self):
        """
        Governs how a tensor is transformed from partial to full 
        (4d tensor) representation.
        """
        if not self.collapsed:
            return self
        T = np.zeros((3, 3, 3, 3), dtype=self._array.dtype)
        m = self._array
        imap = self.imap()
        for ij, ijkl in imap.items():
            T[ijkl] = m[ij]
        self._array = T
        self.collapsed = False
        return self

    def collapse(self):
        """
        Governs how a tensor is transformed from full (4d tensor) to
        partial representation.
        """
        if self.collapsed:
            return self
        m = np.zeros((6, 6), dtype=self._array.dtype)
        T = self._array
        imap = self.imap()
        for ij, ijkl in imap.items():
            m[ij] = T[ijkl]
        self._array = m
        self.collapsed = True
        return self

    @classmethod
    def imap(cls, imap1d=None, *args, **kwargs):
        """
        Returns a 2d-to-4d index map used to collapse or expand a tensor,
        based on the 1d-to-2d mapping of the class the function is called on,
        or on the first argument, if it is a suitable candidate for an
        index map.
        """
        if imap1d is None:
            imap1d = cls.__imap__
        indices = np.indices((6, 6))
        it = np.nditer([*indices], ['multi_index'])
        imap2d = dict()
        for _ in it:
            i, j = it.multi_index
            imap2d[(i, j)] = imap1d[i] + imap1d[j]
        return imap2d

    @classmethod
    def symbolic(cls, *args, base='C_', as_matrix=False, imap=None,
                 **kwargs):
        """
        Returns a symbolic representation of a 4th order 3x3x3x3 tensor.
        If the argument 'as_matrix' is True, the function returns a 6x6 matrix,
        that unfolds according to the argument 'imap', or if it's not provided,
        the index map of the class the function is called on. If 'imap' is
        provided, it must be a dictionary including exactly 6 keys and
        values. The keys must be integers in the integer range (0, 6), the
        values must be tuples on the integer range (0, 3).
        The default mapping is

                0 : (0, 0) --> normal stress x
                1 : (1, 1) --> normal stress y
                2 : (2, 2) --> normal stress z
                3 : (1, 2) --> shear stress yz
                4 : (0, 2) --> shear stress xz
                5 : (0, 1) --> shear stress xy

        and it means the classical Voigt unfolding of the tensor indices.
        """
        res = np.zeros((3, 3, 3, 3), dtype=object)
        indices = np.indices((3, 3, 3, 3))
        it = np.nditer([*indices], ['multi_index'])
        for _ in it:
            p, q, r, s = it.multi_index
            if q >= p and s >= r:
                sinds = np.array([p, q, r, s], dtype=np.int16) + 1
                sym = sy.symbols(base + '_'.join(sinds.astype(str)))
                res[p, q, r, s] = sym
                res[q, p, r, s] = sym
                res[p, q, s, r] = sym
                res[q, p, s, r] = sym
                res[r, s, p, q] = sym
                res[r, s, q, p] = sym
                res[s, r, p, q] = sym
                res[s, r, q, p] = sym
        if as_matrix:
            mat = np.zeros((6, 6), dtype=object)
            imap = cls.imap(imap) if imap is None else imap
            for ij, ijkl in imap.items():
                mat[ij] = res[ijkl]
            if 'sympy' in args:
                res = sy.Matrix(mat)
            else:
                res = mat
        return res

    def _transform(self, dcm: np.ndarray):
        """
        Returns the components of the transformed numerical tensor, based on
        the provided direction cosine matrix.
        """
        if self.collapsed:
            self.expand()
            array = tr_3333_jit(self._array, dcm)
            self.collapse()
        else:
            array = tr_3333_jit(self._array, dcm)
        return array

    def _transform_sym(self, dcm: np.ndarray):
        """
        Returns the components of the transformed symbolic tensor, based on
        the provided direction cosine matrix.
        """
        if self.collapsed:
            self.expand()
            array = tr_3333(self._array, dcm, dtype=object)
            self.collapse()
        else:
            array = tr_3333(self._array, dcm, dtype=object)
        return array


class ComplianceTensor(Tensor3333):

    def __init__(self, *args, imap=None, **kwargs):
        super().__init__(*args, imap=imap, **kwargs)


if __name__ == '__main__':
    from dewloosh.math.linalg.frame import ReferenceFrame

    A = ReferenceFrame()
    B = A.orient_new('Body', [0, 0, 90*np.pi/180], 'XYZ')

    tA = Tensor([[1, 0, 0], [0, 0, 0], [0, 0, 0]], frame=A)
    tA.transform_to_frame(B)

    C = Tensor3333(np.zeros((3, 3, 3, 3)), frame=A)
    C.collapse()
    C.orient(B)


# -*- coding: utf-8 -*-
import numpy as np
from collections import namedtuple


__all__ = ['Quadrature', 'GaussPoints', 'GaussPoints1D', 
           'GaussPoints2D', 'GaussPoints3D']


Quadrature = namedtuple('QuadratureRule', ['inds', 'pos', 'weight'])


def GaussPoints(*args):
    nD = len(args)
    if nD == 1:
        return GaussPoints1D(args[0])
    elif nD == 2:
        return GaussPoints2D(args)
    elif nD == 3:
        return GaussPoints3D(args)


def GaussPoints1D(NumPoints):
    x, w = np.polynomial.legendre.leggauss(NumPoints)
    v = np.zeros([2, NumPoints])
    v[0, :] = x
    v[1, :] = w
    return v


def GaussPoints2D(NumPoints):
    nGaus = NumPoints[0] * NumPoints[1]
    QuadraturePos = np.zeros((nGaus, 2))
    QuadratureWeight = np.zeros((nGaus))
    quad1 = GaussPoints1D(NumPoints[0])
    quad2 = GaussPoints1D(NumPoints[1])
    g = 0
    for gi in np.nditer(quad1, flags=['external_loop'], order='F'):
        for gj in np.nditer(quad2, flags=['external_loop'], order='F'):
            QuadraturePos[g] = np.array([gi[0], gj[0]])
            QuadratureWeight[g] = gi[1] * gj[1]
            g += 1
    return QuadraturePos, QuadratureWeight


def GaussPoints3D(NumPoints):
    nGaus = NumPoints[0] * NumPoints[1] * NumPoints[2]
    QuadraturePos = np.zeros((nGaus, 3))
    QuadratureWeight = np.zeros((nGaus))
    quad1 = GaussPoints1D(NumPoints[0])
    quad2 = GaussPoints1D(NumPoints[1])
    quad3 = GaussPoints1D(NumPoints[2])
    g = 0
    for gi in np.nditer(quad1, flags=['external_loop'], order='F'):
        for gj in np.nditer(quad2, flags=['external_loop'], order='F'):
            for gk in np.nditer(quad3, flags=['external_loop'], order='F'):
                QuadraturePos[g] = np.array([gi[0], gj[0], gk[0]])
                QuadratureWeight[g] = gi[1] * gj[1] * gk[1]
                g += 1
    return QuadraturePos, QuadratureWeight



# -*- coding: utf-8 -*-
from typing import TypeVar, Callable
import sympy as sy
from sympy import Expr, degree, symbols, latex, lambdify
from sympy.core.numbers import One
import numpy as np
from collections import OrderedDict

from dewloosh.core.tools import getasany

from .meta import MetaFunction, substitute


__all__ = ['Function', 'VariableManager', 'FuncionLike']


FuncionLike = TypeVar('FuncionLike', str, Callable, Expr)


class Function(MetaFunction):
    """
    Base class function.
    Initializes a base class for further operation.
    Possible arguments for **kwargs:
        value
        gradient
        Hessian
        domain
    """

    def __init__(self, f0: FuncionLike = None, f1: Callable = None,
                 f2: Callable = None, *args, **kwargs):
        super().__init__()
        self.update(f0, f1, f2, *args, **kwargs)

    def update(self, f0: FuncionLike = None, f1: Callable = None,
               f2: Callable = None, *args, **kwargs):
        self.from_str = None
        if f0 is not None:
            if isinstance(f0, str):
                kwargs.update(self._str_to_func(f0, **kwargs))
                self.from_str = True
            elif isinstance(f0, Expr):
                kwargs.update(self._sympy_to_func(f0, **kwargs))

        self.expr = kwargs.get('expr', None)
        self.variables = kwargs.get('variables', None)
        self.f0 = kwargs.get('value', f0)
        self.f1 = kwargs.get('gradient', f1)
        self.f2 = kwargs.get('Hessian', f2)
        self.dimension = getasany(['d', 'dimension', 'dim'], None, **kwargs)
        self.domain = kwargs.get('domain', None)
        self.vmap = kwargs.get('vmap', None)

    @property
    def symbolic(self):
        try:
            return self.expr is not None
        except AttributeError:
            return False

    @property
    def linear(self):
        if self.symbolic:
            return all(np.array([degree(self.expr, v)
                                 for v in self.variables], dtype=int) == 1)
        else:
            try:
                G = self.G().astype(np.int32)
                assert np.all((G == 0))
                return True
            except AssertionError:
                return False

    def linear_coefficients(self, normalize=False):
        d = self.coefficients(normalize)
        if d:
            return {key: value for key, value in d.items()
                    if len(key.free_symbols) <= 1}
        return None

    def coefficients(self, normalize=False):
        try:
            d = OrderedDict({x: 0 for x in self.variables})
            d.update(self.expr.as_coefficients_dict())
            if not normalize:
                return d
            else:
                res = OrderedDict()
                for key, value in d.items():
                    if len(key.free_symbols) == 0:
                        res[One()] = value*key
                    else:
                        res[key] = value
                return res
        except Exception:
            return None

    def to_latex(self):
        try:
            return latex(self.expr)
        except Exception:
            return None

    def subs(self, values, variables=None, inplace=False):
        if self.expr is None:
            return None
        expr = substitute(self.expr, values, variables,
                          as_string=self.from_str)
        kwargs = self._sympy_to_func(expr=expr, variables=variables)
        if not inplace:
            return Function(None, None, None, **kwargs)
        else:
            self.update(None, None, None, **kwargs)
            return self


class VariableManager(object):

    def __init__(self, variables=None, **kwargs):
        try:
            variables = list(sy.symbols(variables, **kwargs))
        except Exception:
            variables = variables
        try:
            self.vmap = OrderedDict({v: v for v in variables})
        except Exception:
            self.vmap = OrderedDict()

    def substitute(self, vmap: dict = None, inverse=False):
        if not inverse:
            sval = list(vmap.values())
            svar = list(vmap.keys())
        else:
            sval = list(vmap.keys())
            svar = list(vmap.values())
        for v, expr in self.vmap.items():
            self.vmap[v] = substitute(expr, sval, svar)

    def lambdify(self, variables=None):
        assert variables is not None
        for v, expr in self.vmap.items():
            self.vmap[v] = lambdify([variables], expr, 'numpy')

    def __call__(self, v):
        return self.vmap[v]

    def target(self):
        return list(self.vmap.keys())

    def source(self):
        s = set()
        for expr in self.vmap.values():
            s.update(expr.free_symbols)
        return list(s)

    def add_variables(self, variables, overwrite=True):
        if overwrite:
            self.vmap.update({v: v for v in variables})
        else:
            for v in variables:
                if v not in self.vmap:
                    self.vmap[v] = v


if __name__ == '__main__':

    def f0(x=None, y=None):
        return x**2 + y

    def f1(x=None, y=None):
        return np.array([2*x, 1])

    def f2(x=None, y=None):
        return np.array([[0, 0], [0, 0]])

    f = Function(f0, f1, f2, d=2)
    print(f.linear)

    g = Function('3*x + 4*y - 2', variables=['x', 'y', 'z'])
    print(g([1, 2, 30]))

    h = Function('3*x + 4*y - 2')
    print(h([1, 2]))

    m = Function('3*x + 4*y - 2', variables=symbols('x y z'))
    print(m([1, 2, -30]))

    x1, x3 = sy.symbols(['x1', 'x3'], positive=False)
    x2, x4 = sy.symbols(['x2', 'x4'], positive=True)
    variables = [x1, x2, x3, x4]
    mngr = VariableManager(variables)
    y1, y2 = sy.symbols(['y1', 'y2'])
    mngr.substitute({y1: 1, y2: 2})



# -*- coding: utf-8 -*-
from numba.extending import unbox, NativeValue
from numba.extending import box
from numba.core import cgutils
from numba.extending import lower_builtin
from numba.extending import overload_attribute, overload
from numba.extending import make_attribute_wrapper
from numba.extending import models, register_model
from numba.extending import type_callable
from numba.extending import typeof_impl
from numba import types
import numpy as np
import operator


class Interval(object):
    """
    A half-open interval on the real number line.
    """

    def __init__(self, lo, hi, arr):
        self.lo = lo
        self.hi = hi
        self._arr = arr

    def __repr__(self):
        return 'Interval(%f, %f)' % (self.lo, self.hi)

    @property
    def width(self):
        return self.hi - self.lo

    def mean(self):
        return np.mean(self._arr)

    def __getitem__(self, key):
        return self._arr[key]


class IntervalType(types.Type):
    def __init__(self):
        self.data = types.Array(types.float64, 1, 'C')
        super(IntervalType, self).__init__(name='Interval')


make_attribute_wrapper(IntervalType, 'lo', 'lo')
make_attribute_wrapper(IntervalType, 'hi', 'hi')
make_attribute_wrapper(IntervalType, 'data', 'data')


@typeof_impl.register(Interval)
def typeof_index(val, c):
    return IntervalType()


@type_callable(Interval)
def type_interval(context):
    def typer(lo, hi, data):
        if isinstance(lo, types.Float) and isinstance(hi, types.Float):
            return IntervalType()
    return typer


@register_model(IntervalType)
class StructModel(models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ('lo', types.float64),
            ('hi', types.float64),
            ('data', fe_type.data),
        ]
        models.StructModel.__init__(self, dmm, fe_type, members)


@overload_attribute(IntervalType, "width")
def get_width(interval):
    def getter(interval):
        return interval.hi - interval.lo
    return getter


@overload_attribute(IntervalType, "mean")
def get_mean(interval):
    def getter(interval):
        return np.mean(interval.data)
    return getter


@overload(operator.getitem)
def overload_dummy_getitem(obj, idx):
    if isinstance(obj, IntervalType):

        def dummy_getitem_impl(obj, idx):
            return obj.data[idx]

        return dummy_getitem_impl


@lower_builtin(Interval, types.Float, types.Float, types.Array(types.float64, 1, 'C'))
def impl_interval(context, builder, sig, args):
    typ = sig.return_type
    lo, hi, data = args
    interval = cgutils.create_struct_proxy(typ)(context, builder)
    interval.lo = lo
    interval.hi = hi
    interval.data = data
    return interval._getvalue()


@unbox(IntervalType)
def unbox_interval(typ, obj, c):
    """
    Convert a Interval object to a native interval structure.
    """
    lo_obj = c.pyapi.object_getattr_string(obj, "lo")
    hi_obj = c.pyapi.object_getattr_string(obj, "hi")
    data_obj = c.pyapi.object_getattr_string(obj, "_arr")
    interval = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    interval.lo = c.pyapi.float_as_double(lo_obj)
    interval.hi = c.pyapi.float_as_double(hi_obj)
    interval.data = c.unbox(typ.data, data_obj).value
    c.pyapi.decref(lo_obj)
    c.pyapi.decref(hi_obj)
    c.pyapi.decref(data_obj)
    is_error = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(interval._getvalue(), is_error=is_error)


@box(IntervalType)
def box_interval(typ, val, c):
    """
    Convert a native interval structure to an Interval object.
    """
    interval = cgutils.create_struct_proxy(
        typ)(c.context, c.builder, value=val)
    lo_obj = c.pyapi.float_from_double(interval.lo)
    hi_obj = c.pyapi.float_from_double(interval.hi)
    data_obj = c.box(typ.data, interval.data)
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(Interval))
    res = c.pyapi.call_function_objargs(class_obj, (lo_obj, hi_obj, data_obj))
    c.pyapi.decref(lo_obj)
    c.pyapi.decref(hi_obj)
    c.pyapi.decref(data_obj)
    c.pyapi.decref(class_obj)
    return res


if __name__ == '__main__':
    from numba import jit

    @jit(nopython=True)
    def inside_interval(interval, x):
        return interval.lo <= x < interval.hi

    @jit(nopython=True)
    def interval_width(interval):
        return interval.width

    @jit(nopython=True)
    def interval_data(interval):
        return interval.data

    @jit(nopython=True)
    def interval_getitem(interval, i):
        return interval[i]

    @jit(nopython=True)
    def new_interval(lo, hi, data):
        return Interval(lo, hi, data)

    lo = 1.0
    hi = 3.0
    data = np.array([1.1, 3.1, 2.1])
    print(new_interval(lo, hi, data)._arr)
    print(interval_data(new_interval(lo, hi, data)))
    print(interval_getitem(new_interval(lo, hi, data), 0))



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray

from .utils import show_vector, show_vectors
from .frame import ReferenceFrame as Frame
from .array import ArrayBase, Array


__all__ = ['Vector']


class VectorBase(ArrayBase):

    def __new__(subtype, shape=None, dtype=float, buffer=None,
                offset=0, strides=None, order=None, frame=None):
        obj = super().__new__(subtype, shape, dtype, buffer,
                              offset, strides, order)
        obj._frame = frame
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self._frame = getattr(obj, '_frame', None)

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, value):
        if isinstance(value, Frame):
            self._frame = value
        else:
            raise TypeError('Value must be a {} instance'.format(Frame))


class Vector(Array):

    _array_cls_ = VectorBase
    _frame_cls_ = Frame

    def __init__(self, *args, frame=None, **kwargs):
        cls_params = kwargs.get('cls_params', dict())
        if frame is not None:
            cls_params['frame'] = frame
        kwargs['cls_params'] = cls_params
        super().__init__(*args, **kwargs)
        if self._array._frame is None:
            self._array._frame = self._frame_cls_(dim=self._array.shape)

    @property
    def array(self) -> VectorBase:
        return self._array

    @array.setter
    def array(self, value):
        buf = np.array(value)
        assert buf.shape == self._array.shape
        self._array = self._array_cls_(shape=buf.shape, buffer=buf,
                                       dtype=buf.dtype)

    def show(self, target: Frame = None, *args, dcm=None, **kwargs) -> ndarray:
        """
        Returns the components in a target frame. If the target is None,
        returns the components in the global frame.

        Returns:
        --------        
        numpy.ndarray

        """
        if not isinstance(dcm, ndarray):
            if target is None:
                target = Frame(dim=self._array.shape[-1])
            dcm = self.frame.dcm(target=target)
        if len(self.array.shape) == 1:
            return show_vector(dcm, self.array)  # dcm @ arr  
        else:
            return show_vectors(dcm, self.array)  # dcm @ arr 
        
    def orient(self, *args, **kwargs):
        dcm = Frame.eye(dim=len(self)).orient_new(*args, **kwargs).dcm()
        self.array = dcm.T @ self._array
        return self

    def orient_new(self, *args, keep_frame=True, **kwargs):
        dcm = Frame.eye(dim=len(self)).orient_new(*args, **kwargs).dcm()
        if keep_frame:
            array = dcm.T @ self._array
            return Vector(array, frame=self.frame)
        else:
            raise NotImplementedError

    def __repr__(self):
        return np.ndarray.__repr__(self._array)

    def __str__(self):
        return np.ndarray.__str__(self._array)


if __name__ == '__main__':

    A = Frame(dim=3)
    B = A.orient_new('Body', [0, 0, 30*np.pi/180], 'XYZ')
    C = B.orient_new('Body', [0, 0, 30*np.pi/180], 'XYZ')

    vA = Vector([1.0, 1.0, 0.0], frame=A)
    vB = vA.orient_new('Body', [0, 0, -30*np.pi/180], 'XYZ')
    vC = Vector(vA.show(C), frame=C)



# -*- coding: utf-8 -*-
from abc import abstractmethod
import numpy as np


__all__ = ['BinaryGeneticAlgorithm']


def even(n): return n % 2 == 0
def odd(n): return not even(n)


class GeneticAlgorithm:
    """
    Binary Generic algorithm for continuous problems.

    positional arguments:
        fnc : Callable
        ranges : sequence of pairs of limits for each variable
        length : chromosome length / string length
        p_c : probability of crossover
        p_m : probability of mutation
        nPop : number of members in the population
        elitism : float or integer
    """

    def __init__(self, fnc, ranges, length=5, p_c=1, p_m=0.2, nPop=100,
                 *args, **kwargs):
        super().__init__()
        self.fnc = fnc
        self.ranges = np.array(ranges)
        self.dim = fnc.dimension
        self.length = length
        self.p_c = p_c
        self.p_m = p_m

        # Second half of the population is used as a pool to make parents.
        # This assumes that population size is a multiple of 4.
        if odd(nPop):
            nPop += 1
        if odd(int(nPop/2)):
            nPop += 2
        assert nPop % 4 == 0
        assert nPop >= 4

        self.nPop = nPop
        self._genotypes = None
        self._fittness = None
        self.reset()
        self.set_solution_params(**kwargs)

    @property
    def genotypes(self):
        return self._genotypes

    @genotypes.setter
    def genotypes(self, value):
        self._genotypes = value
        self.phenotypes = self.decode(self._genotypes)

    def reset(self):
        self._evolver = self.evolver()
        self._evolver.send(None)

    def set_solution_params(self, tol=1e-12, maxiter=200, miniter=100,
                            elitism=1, **kwargs):
        self.tol = tol
        self.maxiter = np.max([miniter, maxiter])
        self.miniter = np.min([miniter, maxiter])
        self.elitism = elitism

    def evolver(self):
        self.genotypes = self.populate()
        _ = yield
        while True:
            self.genotypes = self.populate(
                self.select(self._genotypes, self.phenotypes))
            yield self._genotypes

    def evolve(self, cycles=1):
        for _ in range(cycles):
            next(self._evolver)
        return self.genotypes

    def criteria(self) -> bool:
        value = yield
        while True:
            _value = yield
            yield abs(value - _value) < self.tol
            value = _value

    def solve(self, reset=False, returnlast=False, **kwargs):
        if reset:
            self.reset()
        self.set_solution_params(**kwargs)
        criteria = self.criteria()
        criteria.send(None)
        criteria.send(self.fnc(self.best_phenotype()))
        finished = False
        nIter = 0
        while (not finished and nIter < self.maxiter) or \
                (nIter < self.miniter):
            next(self._evolver)
            finished = criteria.send(self.fnc(self.best_phenotype()))
            next(criteria)
            nIter += 1
        self.nIter = nIter
        return self.best_phenotype(lastknown=returnlast)

    def fittness(self, phenotypes=None, dtype=np.float32):
        if phenotypes is not None:
            self._fittness = np.array([self.fnc(x) for x in phenotypes],
                                      dtype=dtype)
        return self._fittness.astype(dtype)

    def best_phenotype(self, lastknown=False):
        if lastknown:
            fittness = self._fittness
        else:
            fittness = self.fittness(self.phenotypes)
        best = np.argmin(fittness)
        return self.phenotypes[best]

    def divide(self, fittness=None):
        """
        Divides population to elit and others,
        and returns the corresponding index arrays.
        """
        if self.elitism < 1:
            argsort = np.argsort(fittness)
            elit = argsort[:int(self.nPop*self.elitism)]
            others = argsort[int(self.nPop*self.elitism):]
        elif self.elitism > 1.1:
            argsort = np.argsort(fittness)
            elit = argsort[:self.elitism]
            others = argsort[self.elitism:]
        else:
            elit = []
            others = list(range(self.nPop))
        return list(elit), others

    def random_parents_generator(self, genotypes=None):
        """
        Returns random pairs from a list of genotypes.
        This assumes theat the length of the input array is
        a multiple of 2.
        """
        n = len(genotypes)
        assert n % 2 == 0
        pool = np.full(n, True)
        nPool = n
        while nPool > 2:
            where = np.argwhere(pool == True).flatten()
            nPool = len(where)
            pair = np.random.choice(where, 2, replace=False)
            parent1 = genotypes[pair[0]]
            parent2 = genotypes[pair[1]]
            pool[pair] = False
            yield parent1, parent2

    @abstractmethod
    def populate(self, genotypes=None):
        ...

    @abstractmethod
    def decode(self, genotypes=None):
        ...

    @abstractmethod
    def crossover(self, parent1=None, parent2=None):
        ...

    @abstractmethod
    def mutate(self, child=None):
        ...

    @abstractmethod
    def select(self, genotypes=None, phenotypes=None):
        ...


class BinaryGeneticAlgorithm(GeneticAlgorithm):

    def populate(self, genotypes=None):
        nPop = self.nPop
        if genotypes is None:
            poolshape = (int(nPop / 2), self.dim * self.length)
            genotypes = np.random.randint(2, size=poolshape)
        else:
            poolshape = genotypes.shape
        nParent = poolshape[0]
        if nParent < nPop:
            offspring = []
            g = self.random_parents_generator(genotypes)
            try:
                while (len(offspring) + nParent) < nPop:
                    parent1, parent2 = next(g)
                    offspring.extend(self.crossover(parent1, parent2))
                genotypes = np.vstack([genotypes, offspring])
            except Exception:
                raise RuntimeError
        return genotypes

    def decode(self, genotypes=None):
        span = (2**self.length - 2**0)
        genotypes = genotypes.reshape((self.nPop, self.dim, self.length))
        precisions = [(self.ranges[d, -1] - self.ranges[d, 0]) / span
                      for d in range(self.dim)]
        phenotypes = \
            np.sum([genotypes[:, :, i]*2**i
                    for i in range(self.length)], axis=0).astype(np.float32)
        for d in range(self.dim):
            phenotypes[:, d] *= precisions[d]
            phenotypes[:, d] += self.ranges[d, 0]
        return phenotypes

    def crossover(self, parent1=None, parent2=None, nCut=None):
        if np.random.rand() > self.p_c:
            return parent1, parent2

        if nCut is None:
            nCut = np.random.randint(1, self.dim*self.length-1)

        cuts = [0, self.dim * self.length]
        p = np.random.choice(range(1, self.length * self.dim - 1),
                             nCut, replace=False)
        cuts.extend(p)
        cuts = np.sort(cuts)

        child1 = np.zeros(self.dim*self.length, dtype=np.int32)
        child2 = np.zeros(self.dim*self.length, dtype=np.int32)

        randBool = np.random.rand() > 0.5
        for i in range(nCut+1):
            if (i % 2 == 0) == randBool:
                child1[cuts[i]:cuts[i+1]] = parent1[cuts[i]:cuts[i+1]]
                child2[cuts[i]:cuts[i+1]] = parent2[cuts[i]:cuts[i+1]]
            else:
                child1[cuts[i]:cuts[i+1]] = parent2[cuts[i]:cuts[i+1]]
                child2[cuts[i]:cuts[i+1]] = parent1[cuts[i]:cuts[i+1]]

        return self.mutate(child1), self.mutate(child2)

    def mutate(self, child=None):
        p = np.random.rand(self.dim*self.length)
        return np.where(p > self.p_m, child, 1-child)

    def select(self, genotypes=None, phenotypes=None):
        fittness = self.fittness(phenotypes)
        winners, others = self.divide(fittness)
        while len(winners) < int(self.nPop / 2):
            candidates = np.random.choice(others, 3, replace=False)
            winner = np.argsort([fittness[ID] for ID in candidates])[0]
            winners.append(candidates[winner])
        return np.array([genotypes[w] for w in winners], dtype=np.float32)


if __name__ == '__main__':

    def Rosenbrock(a, b, x, y):
        return (a-x)**2 + b*(y-x**2)**2

    def f(x):
        return Rosenbrock(1, 100, x[0], x[1])

    f.dimension = 2
    ranges = [
        [-10, 10],
        [-10, 10]
    ]
    BGA = BinaryGeneticAlgorithm(f, ranges, length=12, nPop=200)
    r = BGA.solve()
    print(r)
    
    # plot the history
    import matplotlib.pyplot as plt
    BGA = BinaryGeneticAlgorithm(f, ranges, length=12, nPop=200)
    history = [f(BGA.best_phenotype())]
    for _ in range(100):
        BGA.evolve(1)
        history.append(f(BGA.best_phenotype()))
    plt.plot(history)
    plt.show()
    x = BGA.best_phenotype()
    fx = f(x)
    print('min {} @ {}'.format(fx,x))
    



# -*- coding: utf-8 -*-
import sympy as sy
from .testfunction import TestMinFunction
from .meta import symbolize, substitute


def Rosenbrock(a=1, b=100):
    """
    Creates the Rosenbrock function object.

    f(x,y) = (a-x)**2 + b*(y-x**2)**2

    a,b are constants, tipically a is set to 1 and  b is set to 100.

    The global minimum
    f(1,1) = 0.
    """
    str_expr = '(a-x)**2 + b*(y-x**2)**2'
    expr = substitute(sy.sympify(str_expr), [a, b], ['a', 'b'])
    return TestMinFunction(**symbolize(expr=expr), optimums=[(1.0, 1.0)])


def Himmelblau():
    """
    Creates the Himmelblau's function object.

    f(x,y) = (x**2 + y - 11)**2 + (x + y**2 - 7)**2

    It has four identical local minima with values 0.0 at
    (3.0,2.0),
    (-2.805,3.131),
    (-3.779,-3.283),
    (3.584,-1.848).
    """
    str_expr = '(x**2 + y - 11)**2 + (x + y**2 - 7)**2'
    return TestMinFunction(**symbolize(str_expr=str_expr),
                           optimums=[(3.0, 2.0), (-2.805118, 3.131312),
                                     (-3.779310, -3.283186),
                                     (3.584428, -1.848126)])


def GoldsteinPrice():
    """
    Creates the Goldstein-Price function object.

    f(x,y) = (1+(x+y+1)**2 * (19-14*x+3*x**2-14*y+6*x*y+3*y**2))
    *(30+(2*x-3*y)**2 * (18-32*x+12*x**2+48*y-36*x*y+27*y**2))

    The global minimum
    f(0,-1) = 3.
    """
    str_expr = '(1+(x+y+1)**2 * (19-14*x+3*x**2-14*y+6*x*y+3*y**2))' \
        '*(30+(2*x-3*y)**2 * (18-32*x+12*x**2+48*y-36*x*y+27*y**2))'
    return TestMinFunction(**symbolize(str_expr=str_expr),
                           optimums=[(0.0, -1.0)])


def Beale():
    """
    Creates the Beale function object.

    f(x,y) = (1.5 - x + x*y)**2 + (2.25 - x + x*y**2)**2
    + (2.625 - x + x*y**3)**2

    The global minimum
    f(3,0.5) = 0.
    """
    str_expr = '(1.5 - x + x*y)**2 + (2.25 - x + x*y**2)**2 ' \
        '+ (2.625 - x + x*y**3)**2'
    return TestMinFunction(**symbolize(str_expr=str_expr),
                           optimums=[(3.0, 0.5)])


def Matyas():
    """
    Creates the Matyas function object.

    f(x, y) = 0.26*(x**2 + y**2) - 0.48*x*y

    The global minimum
    f(0.,0.) = 0.
    """
    str_expr = '0.26*(x**2 + y**2) - 0.48*x*y'
    return TestMinFunction(**symbolize(str_expr=str_expr),
                           optimums=[(0.0, 0.0)])


if __name__ == '__main__':

    """
    check:
        - Himmelblau with initial = [-1.,-1.]. Problem with the Hessian
          at [-0.270845, -0.923039].

    """



# -*- coding: utf-8 -*-
import numpy as np
from numpy.linalg import LinAlgError
from numba import njit, guvectorize
__cache = True


__all__ = ['solve', 'reduce', 'backsub', 'npsolve', 'inv3x3']


def solve(A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray = None,
          presc_val: np.ndarray = None, method='numpy', inplace=False):
    if method == 'numpy':
        assert presc_bool is None, \
            "Method '{}' does not support prescribed values.".format(method)
        return npsolve(A, B)
    elif method in ['Jordan', 'Gauss-Jordan']:
        fnc = _Jordan if method == 'Jordan' else _GaussJordan
        try:
            nEQ = len(B)
            if len(B.shape) == 1:
                B = B.reshape((nEQ, 1))
            pre = presc_val is not None
            if not pre:
                presc_bool = np.zeros((nEQ,), dtype=np.int32)
                presc_val = np.zeros((nEQ,), dtype=np.float32)
            if inplace:
                res = backsub(*fnc(A, B, presc_bool, presc_val),
                              presc_bool, presc_val)
            else:
                res = backsub(*fnc(A.copy(), B.copy(), presc_bool, presc_val),
                              presc_bool, presc_val)
            if pre:
                return res
            else:
                return res[0]
        except Exception:
            raise LinAlgError('The matrix is singular!')
    else:
        raise AttributeError("Invalid method name '{}'.".format(method))


def reduce(A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray = None,
           presc_val: np.ndarray = None, method='Gauss-Jordan',
           inplace=False):
    fnc = None
    if method == 'Gauss-Jordan':
        fnc = _GaussJordan
    elif method == 'Jordan':
        fnc = _Jordan
    if fnc is not None:
        try:
            if presc_bool is None:
                nEQ = len(B)
                presc_bool = np.zeros((nEQ,), dtype=np.int32)
                presc_val = np.zeros((nEQ,), dtype=np.float32)
            if inplace:
                return fnc(A, B, presc_bool, presc_val)
            else:
                return fnc(A.copy(), B.copy(), presc_bool, presc_val)
        except Exception:
            raise LinAlgError('The matrix is singular!')
    else:
        raise AttributeError("Invalid method name '{}'.".format(method))


@njit(nogil=True, cache=__cache)
def npsolve(A, b):
    return np.linalg.solve(A, b)


@guvectorize(['(f8[:, :], f8[:, :])', '(f4[:, :], f4[:, :])'],
             '(n, n) -> (n, n)', nopython=True, cache=__cache)
def inv3x3(A, res):
    d = A[0, 0] * A[1, 1] * A[2, 2] - A[0, 0] * A[1, 2] * A[2, 1] \
        - A[0, 1] * A[1, 0] * A[2, 2] + A[0, 1] * A[1, 2] * A[2, 0] \
        + A[0, 2] * A[1, 0] * A[2, 1] - A[0, 2] * A[1, 1] * A[2, 0]
    res[0, 0] = A[1, 1] * A[2, 2] - A[1, 2] * A[2, 1]
    res[0, 1] = -A[0, 1] * A[2, 2] + A[0, 2] * A[2, 1]
    res[0, 2] = A[0, 1] * A[1, 2] - A[0, 2] * A[1, 1]
    res[1, 0] = -A[1, 0] * A[2, 2] + A[1, 2] * A[2, 0]
    res[1, 1] = A[0, 0] * A[2, 2] - A[0, 2] * A[2, 0]
    res[1, 2] = -A[0, 0] * A[1, 2] + A[0, 2] * A[1, 0]
    res[2, 0] = A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]
    res[2, 1] = -A[0, 0] * A[2, 1] + A[0, 1] * A[2, 0]
    res[0, 0] = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res /= d


@njit(nogil=True, cache=__cache)
def _Jordan(A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray,
            presc_val: np.ndarray):
    nEQ = len(B)
    for iEQ in range(nEQ):
        if presc_bool[iEQ] == True:
            for iROW in range(iEQ, nEQ):
                B[iROW] -= A[iROW, iEQ] * presc_val[iEQ]
                A[iROW, iEQ] = 0.0
        else:
            pivot = A[iEQ, iEQ]
            if abs(pivot) < 1e-12:
                raise Exception('The matrix is singular!')
            for iROW in range(nEQ):
                factor = A[iROW, iEQ] / pivot
                if iROW == iEQ or abs(factor) < 1e-12:
                    continue
                for iCOL in range(nEQ):
                    A[iROW, iCOL] -= factor * A[iEQ, iCOL]
                B[iROW] -= factor * B[iEQ]
    return A, B


@njit(nogil=True, cache=__cache)
def _GaussJordan(A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray,
                 presc_val: np.ndarray):
    nEQ = len(B)
    for iEQ in range(nEQ):
        if presc_bool[iEQ] == True:
            for iROW in range(iEQ, nEQ):
                B[iROW] -= A[iROW, iEQ] * presc_val[iEQ]
                A[iROW, iEQ] = 0.0
        else:
            if iEQ < nEQ-1:
                pivot = A[iEQ, iEQ]
                if abs(pivot) < 1e-12:
                    raise Exception('The matrix is singular!')
                for iROW in range(iEQ + 1, nEQ):
                    factor = A[iROW, iEQ] / pivot
                    if abs(factor) < 1e-12:
                        continue
                    for iCOL in range(nEQ):
                        A[iROW, iCOL] -= factor * A[iEQ, iCOL]
                    B[iROW] -= factor * B[iEQ]
    return A, B


@njit(nogil=True, cache=__cache)
def backsub(A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray,
            presc_val: np.ndarray):
    nEQ, nRHS = B.shape
    X = np.zeros((nEQ, nRHS), dtype=np.float32)
    R = np.zeros((nEQ, nRHS), dtype=np.float32)
    resid = np.zeros(nRHS, dtype=np.float32)
    for iEQ in range(nEQ - 1, -1, -1):
        pivot = A[iEQ, iEQ]
        resid[:] = B[iEQ]
        if iEQ < nEQ - 1:
            for iCOL in range(iEQ + 1, nEQ):
                resid -= A[iEQ, iCOL] * X[iCOL]
        if presc_bool[iEQ] == True:
            X[iEQ] = presc_val[iEQ]
            R[iEQ] = - resid
        else:
            X[iEQ] = resid / pivot
    return X, R


if __name__ == '__main__':
    from time import time

    A = np.array([[3, 1, 2], [1, 1, 1], [2, 1, 2]], dtype=np.float32)
    B = np.array([11, 6, 10], dtype=np.float32)
    presc_bool = np.full(B.shape, False, dtype=np.bool)
    presc_val = np.zeros(B.shape, dtype=np.float32)

    x_J = solve(A, B, method='Jordan')
    x_GJ = solve(A, B, method='Gauss-Jordan')
    x_np = np.linalg.solve(A, B)
    x_np_jit = solve(A, B, method='numpy')
    A_GJ, b_GJ = reduce(A, B, method='Gauss-Jordan')
    A_J, b_J = reduce(A, B, method='Jordan')

    B = np.stack([B, B], axis=1)
    X_J = solve(A, B, method='Jordan')
    X_GJ = solve(A, B, method='Gauss-Jordan')

    A = np.array([[1, 2, 1, 1], [2, 5, 2, 1], [1, 2, 3, 2], [1, 1, 2, 2]],
                 dtype=np.float32)
    b = np.array([12, 15, 22, 17], dtype=np.float32)
    ifpre = np.array([0, 1, 0, 0], dtype=np.int32)
    fixed = np.array([0, 2, 0, 0], dtype=np.float32)
    A_GJ, b_GJ = reduce(A, b, ifpre, fixed, 'Gauss-Jordan')
    A_J, b_J = reduce(A, b, ifpre, fixed, 'Jordan')
    X_GJ, r_GJ = solve(A, b, ifpre, fixed, 'Gauss-Jordan')
    X_J, r_J = solve(A, b, ifpre, fixed, 'Jordan')

    def test_1(A, b, n=100):
        times = []
        
        # measuring solution 1
        ts = time()
        for i in range(n):
            solve(A, b, method='Jordan')
        te = time()
        times.append((te-ts)*1000)
        
        # measuring solution 2
        ts = time()
        for i in range(n):
            np.linalg.solve(A, b)
        te = time()
        times.append((te-ts)*1000)
        
        # measuring solution 3
        ts = time()
        for i in range(n):
            solve(A, b, method='numpy')
        te = time()
        times.append((te-ts)*1000)
        
        # measuring solution 4
        ts = time()
        for i in range(n):
            solve(A, b, method='Gauss-Jordan')
        te = time()
        times.append((te-ts)*1000)
        
        return times

    A = np.array([[3, 1, 2], [1, 1, 1], [2, 1, 2]], dtype=np.float32)
    b = np.array([11, 6, 10], dtype=np.float32)
    print(test_1(A, b, 1000))



# -*- coding: utf-8 -*-
import numpy as np
from numba import jit
from numba.types import int32, int64, Array
from numba.typed import Dict

from .csr import CSR


try:
    import networkx as ntx
    class Graph(ntx.Graph):

        def adjacency_matrix(self, sparse = False):
            if sparse:
                return CSR(ntx.adjacency_matrix(self))
            else:
                return ntx.adjacency_matrix(self)

        def rooted_level_structure(self, root = 0):
            return rooted_level_structure(CSR(ntx.adjacency_matrix(self)), root)

        def pseudo_peripheral_nodes(self):
            return pseudo_peripheral_nodes(CSR(ntx.adjacency_matrix(self)))
except:
    Graph = None


int32A = Array(int32, 1, 'C')
int64A = Array(int64, 1, 'C')


@jit(nopython = True, nogil = True, fastmath = False, cache = False)
def rooted_level_structure(adj : CSR, root : int = 0) -> Dict:
    """
    Turns a sparse adjacency matrix into a rooted level structure.
    """
    nN = len(adj.indptr) - 1
    rls = Dict.empty(
        key_type = int64,
        value_type = int64A,
    )
    level = 0
    rls[level] = np.array([root], dtype = np.int64)
    nodes = np.zeros(nN, dtype = np.int64)
    nodes[root] = 1
    levelset = np.zeros(nN, dtype = np.int64)
    nE = 1
    while nE < nN:
        levelset[:] = 0
        for node in rls[level]:
            _, neighbours = adj.row(node)
            levelset[neighbours] = 1
        for iN in range(nN):
            if nodes[iN] == 1:
                levelset[iN] = 0
        level += 1
        rls[level] = np.where(levelset == 1)[0]
        nE += len(rls[level])
        for iN in range(nN):
            if levelset[iN] == 1:
                nodes[iN] = 1
    return rls


@jit(nopython = True, nogil = True, fastmath = False, cache = False)
def pseudo_peripheral_nodes(adj : CSR):

    def length_width(RLS):
        length = len(RLS)
        width = 0
        for i in range(length):
            width = max(width, len(RLS[i]))
        return length, width

    RLS = rooted_level_structure(adj, root = 0)
    length, width = length_width(RLS)
    while True:
        nodes = RLS[len(RLS) - 1]
        found = False
        for iN, node in enumerate(nodes):
            iRLS = rooted_level_structure(adj, root = node)
            iL, iW = length_width(iRLS)
            if (iL > length) or (iL == length and iW < width):
                RLS = iRLS
                length = iL
                width = iW
                found = True
        if not found:
            nR = len(RLS[len(RLS) - 1]) + 1
            res = np.zeros(nR, dtype = np.int32)
            res[:-1] = RLS[len(RLS) - 1]
            res[-1] = RLS[0][0]
            return res


if __name__ == '__main__':
    pass
    """
    from pyoneer.mechanics.fem import Uniform3D

    # geometry
    Lx = 100.0
    Ly = 100.0
    Lz = 100.0

    # mesh control
    nx = 21
    ny = 21
    nz = 21

    # assembly
    assembly = Uniform3D((Lx, Ly, Lz), (nx, ny, nz))
    points = assembly.points()

    # connectivity and topology
    conn = assembly.element_node_numbering()
    topo = assembly.element_dof_numbering()
    G = assembly.topology('as_graph', mode = 'node')
    adj = assembly.adjacency(G)
    adj_csr = G.adjacency_matrix(sparse = True)
    rls_csr = rooted_level_structure(adj_csr)
    ppnodes = pseudo_peripheral_nodes(adj_csr)
    conn_csr = assembly.connectivity_matrix(sparse = True)
    """


# -*- coding: utf-8 -*-
import numpy as np
from numba import njit, config, prange


def tr_3333(array: np.ndarray, dcm: np.ndarray):
    Q = dcm.T
    res = np.zeros((3, 3, 3, 3), dtype=array.dtype)
    for p in range(3):
        for q in range(3):
            for r in range(3):
                for s in range(3):
                    for i in range(3):
                        for j in range(3):
                            for k in range(3):
                                for m in range(3):
                                    res[p, q, r, s] += \
                                        Q[p, i] * Q[q, j] * Q[r, k] * \
                                        Q[s, m] * array[i, j, k, m]
    return res


@njit(nogil=True, parallel=True, cache=True)
def tr_3333_jit(array: np.ndarray, dcm: np.ndarray):
    Q = dcm.T
    res = np.zeros((3, 3, 3, 3), dtype=array.dtype)
    for p in prange(3):
        for q in prange(3):
            for r in prange(3):
                for s in prange(3):
                    for i in prange(3):
                        for j in prange(3):
                            for k in prange(3):
                                for m in prange(3):
                                    res[p, q, r, s] += \
                                        Q[p, i] * Q[q, j] * Q[r, k] * \
                                        Q[s, m] * array[i, j, k, m]
    return res


def tr_3333_jit_compiled(array: np.ndarray, dcm: np.ndarray):
    ...


_code = compile("""
@njit(nogil = True, parallel = True)
def tr_3333_jit_compiled(array : np.ndarray, dcm : np.ndarray):
    Q = dcm.T
    res = np.zeros((3, 3, 3, 3), dtype = array.dtype)
    for p in prange(3):
        for q in prange(3):
            for r in prange(3):
                for s in prange(3):
                    for i in prange(3):
                        for j in prange(3):
                            for k in prange(3):
                                for m in prange(3):
                                    res[p, q, r, s] += \
                                        Q[p, i] * Q[q, j] * Q[r, k] * \
                                        Q[s, m] * array[i, j, k, m]
    return res
""", 'tr_3333_jit_compiled', 'exec')
exec(_code)


def tr_3333_np(array: np.ndarray, dcm: np.ndarray,
               optimize='greedy') -> np.ndarray:
    Q = dcm.T
    return np.einsum('pi, qj, rk, sl, ijkl', Q, Q, Q, Q, array,
                     optimize=optimize)


if __name__ == '__main__':
    from dewloosh.math.linalg.frame import ReferenceFrame
    from dewloosh.math.linalg.tensor import Tensor, Tensor3333
    import time

    A = ReferenceFrame()
    B = A.orient_new('Body', [0, 0, 90*np.pi/180],  'XYZ')

    tA = Tensor([[1, 0, 0], [0, 0, 0], [0, 0, 0]], frame=A)
    tA.transform_to_frame(B)

    C = Tensor3333(np.zeros((3, 3, 3, 3)), frame=A)
    dcm = B.dcm()
    array = C._array
    Q = np.transpose(dcm)
    path = np.einsum_path('pi,qj,rk,sl,ijkl', Q, Q, Q, Q, array,
                          optimize='greedy')[0]

    # PERFORMENCE MEASURE
    times = []
    # measuring solution 1
    ts = time.time()
    for i in range(100):
        tr_3333(array, dcm)
    te = time.time()
    times.append((te-ts)*1000)
    
    # measuring solution 2
    ts = time.time()
    for i in range(100):
        tr_3333_jit(array, dcm)
    te = time.time()
    times.append((te-ts)*1000)
    
    # measuring solution 3
    ts = time.time()
    for i in range(100):
        tr_3333_jit_compiled(array, dcm)
    te = time.time()
    times.append((te-ts)*1000)
    
    # measuring solution 4
    ts = time.time()
    for i in range(100):
        tr_3333_np(array, dcm)
    te = time.time()
    times.append((te-ts)*1000)
    
    # measuring solution 5
    ts = time.time()
    for i in range(100):
        tr_3333_np(array, dcm, optimize=path)
    te = time.time()
    times.append((te-ts)*1000)
    print(times)



# -*- coding: utf-8 -*-
from enum import Enum

from .function import Function



class ConstraintType(Enum):
    nan = 0
    equality = 1
    inequality = 2


class ConstraintFunction(Function):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.constraintType = ConstraintType.nan
        for arg in args:
            if arg in ['eq', 'equality']:
                self.constraintType = ConstraintType.equality
                break
            elif arg in ['iq', 'inequality']:
                self.constraintType = ConstraintType.inequality
                break
            elif isinstance(arg, ConstraintType):
                self.constraintType = arg
                break
        assert self.constraintType != ConstraintType.nan


class EqualityConstraint(ConstraintFunction):

    def __init__(self, *args, **kwargs):
        super().__init__('eq', *args, **kwargs)


class InequalityConstraint(ConstraintFunction):

    def __init__(self, *args, **kwargs):
        super().__init__('iq', *args, **kwargs)



class PenaltyType(Enum):
    courant = 1


class PenaltyFunction(EqualityConstraint):
    """
    Penalty function class for equality constraints.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.PenaltyType = PenaltyType.courant

        if 'penalty' in kwargs:
            assert isinstance(kwargs['penalty'], float)
            self.penalty = kwargs['penalty']
        else:
            self.penalty = np.array(1e+12)

    def f(self, *args, **kwargs):
        try:
            if self.PenaltyType == PenaltyType.courant:
                f = self.f0(*args, **kwargs)
                p = self.penalty
                return np.multiply(0.5*p, np.power(f, 2))
        except:
            return None

    def g(self, *args, **kwargs):
        try:
            if self.PenaltyType == PenaltyType.courant:
                f = self.f0(*args, **kwargs)
                g = self.f1(*args, **kwargs)
                p = self.penalty
                return np.multiply(np.multiply(p, f), g)
        except:
            return None

    def G(self, *args, **kwargs):
        try:
            if self.PenaltyType == PenaltyType.courant:
                f = self.f0(*args, **kwargs)
                g = self.f1(*args, **kwargs)
                G = self.f2(*args, **kwargs)
                p = self.penalty
                return np.multiply(p, np.add(np.multiply(f, G), np.outer(g, g)))
        except:
            return None



# -*- coding: utf-8 -*-
import numpy as np
from numba import njit, prange

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def transpose_dcm_multi(dcm: np.ndarray):
    N = dcm.shape[0]
    res = np.zeros_like(dcm)
    for i in prange(N):
        res[i] = dcm[i].T
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def dcm_n(dcm: np.ndarray, n: int):
    res = np.zeros((3 * n, 3 * n), dtype=dcm.dtype)
    for i in prange(n):
        _i = i * 3
        i_ = _i + 3
        res[_i:i_, _i:i_] = dcm
    return res


# -*- coding: utf-8 -*-
from enum import Enum
import operator as op
from typing import TypeVar, Callable

from dewloosh.core.tools import getasany

from .function import Function


__all__ = ['Equality', 'InEquality']


class Relations(Enum):
    eq = '='
    gt = '>'
    ge = '>='
    lt = '<'
    le = '<='

    def to_op(self):
        return _rel_to_op[self]


_rel_to_op = {
    Relations.eq: op.eq,
    Relations.gt: op.gt,
    Relations.ge: op.ge,
    Relations.lt: op.lt,
    Relations.le: op.le
}

RelationType = TypeVar('RelationType', str, Relations, Callable)


class Relation(Function):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.op = None
        self.opfunc = None
        op = getasany(['op', 'operator'], None, **kwargs)
        if op:
            if isinstance(op, str):
                self.op = Relations(op)
            elif isinstance(op, Relations):
                self.op = op
            elif isinstance(op, Callable):
                self.opfunc = op
                self.op = None
        else:
            self.op = Relations.eq
        if op and isinstance(self.op, Relations):
            self.opfunc = self.op.to_op()
        self.slack = 0

    @property
    def operator(self):
        return self.op

    def to_eq(self):
        raise NotImplementedError

    def relate(self, *args, **kwargs):
        return self.opfunc(self.f0(*args, **kwargs), 0)

    def __call__(self, *args, **kwargs):
        return self.opfunc(self.f0(*args, **kwargs), 0)


class Equality(Relation):
    def __init__(self, *args, **kwargs):
        kwargs['op'] = Relations.eq
        super().__init__(*args, **kwargs)

    def to_eq(self):
        return self


class InEquality(Relation):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_eq(self):
        raise


if __name__ == '__main__':

    gt = InEquality('x + y', op='>')
    print(gt([0.0, 0.0]))

    ge = InEquality('x + y', op='>=')
    print(ge([0.0, 0.0]))

    le = InEquality('x + y', op=lambda x, y: x <= y)
    print(le([0.0, 0.0]))

    lt = InEquality('x + y', op=lambda x, y: x < y)
    print(lt([0.0, 0.0]))



# -*- coding: utf-8 -*-
import numpy as np
from numba.core import types as nbtypes, cgutils
from numba.extending import typeof_impl, models, make_attribute_wrapper, \
    register_model, box, unbox, NativeValue, overload_method
from scipy.sparse import issparse
from scipy.sparse import csr_matrix as csr_scipy

from .utils import get_shape_sp


__all__ = ['csr_matrix']


class csr_matrix(object):
    """
    Numba-jittable python class for a sparse matrix in csr format. 

    Parameters
    ----------
    data    
        Contains the non-zero values of the matrix, in the order in which
        they would be encountered if we walked along the rows left to
        right and top to bottom. If this is a csc matrix, the walk
        happens along the columns.

    indices
        The indices of the columns (rows) during the walk.

    indptr
        Stores row (column) boundaries.

    """

    def __init__(self, data: None, indices: np.array = None,
                 indptr: np.array = None, shape: tuple = None):
        self.sptype = csr_scipy
        if issparse(data):
            data = data.tocsr()
            self.data = data.data.astype(np.float64)
            self.indices = data.indices.astype(np.int32)
            self.indptr = data.indptr.astype(np.int32)
            self.shape = data.shape
        elif isinstance(data, np.ndarray) and indices is None:
            sp = self.csr_scipy(data)
            self.data = sp.data.astype(np.float64)
            self.indices = sp.indices.astype(np.int32)
            self.indptr = sp.indptr.astype(np.int32)
            self.shape = sp.shape
        else:
            self.data = np.array(data).astype(np.float64)
            self.indices = np.array(indices).astype(np.int32)
            self.indptr = np.array(indptr).astype(np.int32)
            if shape is None:
                shape = get_shape_sp(indptr)
            self.shape = shape

    def to_scipy(self):
        return csr_scipy((self.data, self.indices, self.indptr), shape=self.shape)
    
    @staticmethod
    def eye(N: int, *args, **kwargs):
        indices = np.arange(N)
        indptr = np.arange(N+1)
        data = np.ones(N, dtype=float)
        return csr_matrix(data=data, indices=indices, indptr=indptr, shape=(N, N))

    def first_nonzero_col(self):
        return self.indices[self.indptr[:-1]]

    def new_rows_per_col(self):
        return np.bincount(self.first_nonzero_col(), minlength=self.shape[1])

    def row(self, i: int = 0):
        """Returns the values and colum indices of the i-th row."""
        return self.data[self.indptr[i]: self.indptr[i+1]], \
            self.indices[self.indptr[i]: self.indptr[i+1]]


class csr_matrix_nb(nbtypes.Type):
    """Numba type for a sparse matrix."""

    def __init__(self, dtype):
        self.dtype = dtype
        self.data = nbtypes.Array(self.dtype, 1, 'C')
        self.indices = nbtypes.Array(nbtypes.int32, 1, 'C')
        self.indptr = nbtypes.Array(nbtypes.int32, 1, 'C')
        self.shape = nbtypes.UniTuple(nbtypes.int64, 2)
        super(csr_matrix_nb, self).__init__('csr_matrix')


@overload_method(csr_matrix_nb, 'row')
def row(csr, i: int):
    if isinstance(csr, csr_matrix_nb):
        def row_impl(csr, i: int):
            return csr.data[csr.indptr[i]:csr.indptr[i+1]], \
                csr.indices[csr.indptr[i]:csr.indptr[i+1]]
        return row_impl


@typeof_impl.register(csr_matrix)
def typeof_csr(val, c):
    data = typeof_impl(val.data, c)
    return csr_matrix_nb(data.dtype)


make_attribute_wrapper(csr_matrix_nb, 'data', 'data')
make_attribute_wrapper(csr_matrix_nb, 'indices', 'indices')
make_attribute_wrapper(csr_matrix_nb, 'indptr', 'indptr')
make_attribute_wrapper(csr_matrix_nb, 'shape', 'shape')


@register_model(csr_matrix_nb)
class csr_model(models.StructModel):
    """Data model for nopython mode."""

    def __init__(self, dmm, fe_type):
        members = [
            ('data', fe_type.data),
            ('indices', fe_type.indices),
            ('indptr', fe_type.indptr),
            ('shape', fe_type.shape)
        ]
        models.StructModel.__init__(self, dmm, fe_type, members)


@unbox(csr_matrix_nb)
def unbox_csr(typ, obj, c):
    """Convert a python object to a numba-native structure."""
    data = c.pyapi.object_getattr_string(obj, "data")
    indices = c.pyapi.object_getattr_string(obj, "indices")
    indptr = c.pyapi.object_getattr_string(obj, "indptr")
    shape = c.pyapi.object_getattr_string(obj, "shape")
    matrix = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    matrix.data = c.unbox(typ.data, data).value
    matrix.indices = c.unbox(typ.indices, indices).value
    matrix.indptr = c.unbox(typ.indptr, indptr).value
    matrix.shape = c.unbox(typ.shape, shape).value
    for att in [data, indices, indptr, shape]:
        c.pyapi.decref(att)
    is_error = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(matrix._getvalue(), is_error=is_error)


@box(csr_matrix_nb)
def box_csr(typ, val, c):
    """Convert a numba-native structure to a python object."""
    matrix = cgutils.create_struct_proxy(typ)(c.context, c.builder,
                                              value=val)
    classobj = c.pyapi.unserialize(
        c.pyapi.serialize_object(csr_matrix))
    data_obj = c.box(typ.data, matrix.data)
    indices_obj = c.box(typ.indices, matrix.indices)
    indptr_obj = c.box(typ.indptr, matrix.indptr)
    shape_obj = c.box(typ.shape, matrix.shape)
    matrix_obj = c.pyapi.call_function_objargs(classobj,
                                               (data_obj, indices_obj,
                                                indptr_obj, shape_obj))
    return matrix_obj


if __name__ == '__main__':
    from numba import njit

    @njit
    def csr_row(csr: csr_matrix, i: int):
        return csr.row(i)
    
    @njit
    def csr_data(csr: csr_matrix):
        return csr.data

    @njit
    def csr_m(dtype=np.float64):
        return csr_matrix_nb(dtype)

    np.random.seed = 0
    mat = csr_scipy(np.random.rand(10, 12) > 0.8, dtype=np.int64)
    print(mat.A)

    csr = csr_matrix(mat)
    sdata, scols = csr_row(csr, 0)
    print(sdata, scols)
    
    e = csr_matrix.eye(3)

    print(csr_data(csr))



# -*- coding: utf-8 -*-
import numpy as np
from numpy.linalg import LinAlgError
import sympy as sy
from sympy.utilities.iterables import multiset_permutations
from copy import copy
from collections import defaultdict, Iterable

from dewloosh.core.tools import getasany

from ..function import Function, InEquality, Equality, VariableManager
from ..function.relation import Relations, Relation
from ..function.meta import coefficients



class LLP(object):
    def __init__(self, *args, constraints=None, variables=None,
                 positive=True, **kwargs):
        super().__init__()
        self.standardform = False
        self.obj = getasany(['obj', 'cost', 'payoff', 'fittness', 'f'],
                            None, **kwargs)
        self.constraints = [] if constraints is None else constraints
        self.vmanager = VariableManager(variables, positive=positive)

    def add_constraint(self, *args, **kwargs):
        if isinstance(args[0], Function):
            if isinstance(args[0], InEquality):
                assert args[0].op in [Relations.ge, Relations.le], \
                    "Only '>=' and '<=' arre allowed!"
            c = args[0]
        else:
            c = Relation(*args, **kwargs)
        self.constraints.append(c)

    @property
    def variables(self):
        return self.vmanager.target()

    def sync_variables(self):
        s = set()
        s.update(self.obj.variables)
        for c in self.constraints:
            s.update(c.variables)
        self.vmanager.add_variables(s)

    def shift_variables(self):
        vmap = dict()
        for v in self.vmanager.source():
            if not v.is_positive:
                s = str(v)
                expr = sum(sy.symbols([s + '__+', s + '__-'], positive=True))
                vmap[v] = expr
        self.vmanager.substitute(vmap)

    def get_system_variables(self):
        s = set()
        s.update(self.obj.variables)
        for c in self.constraints:
            s.update(c.variables)
        return list(s)

    def get_slack_variables(self, template='y_{}'):
        inequalities = list(filter(lambda c: isinstance(c, InEquality),
                            self.constraints))
        n = len(inequalities)
        strlist = [template.format(i) for i in range(n)]
        y = sy.symbols(strlist, positive=True)
        for i in range(n):
            inequalities[i].slack = y[i]
        return y

    def has_standardform(self):
        all_eq = all([isinstance(c, Equality) for c in self.constraints])
        all_pos = all([v.is_positive for v in self.get_system_variables()])
        return all_pos and all_eq

    def as_standard(self, inplace=False):
        # handle variables
        self.sync_variables()
        self.shift_variables()
        self.vmanager.add_variables(self.get_slack_variables())
        v = self.vmanager.source()
        n = len(v)
        x = list(sy.symbols(' '.join(['x_{}'.format(i) for i in range(n)]),
                            positive=True))
        c = list(sy.symbols(' '.join(['c_{}'.format(i) for i in range(n)])))
        x_v = {x_: v_ for x_, v_ in zip(x, v)}
        self.vmanager.substitute(vmap=x_v, inverse=True)
        v.append(1)
        x.append(1)
        c.append(sy.symbols('c'))
        x_c = {x_: c_ for x_, c_ in zip(x, c)}
        template = np.inner(c, x)
        x.pop(-1)
        v.pop(-1)
        vmap = self.vmanager.vmap

        def redefine(fnc):
            expr = fnc.expr.subs([(v, expr) for v, expr in vmap.items()])
            fnc_coeffs = coefficients(expr=expr, normalize=True)
            coeffs = defaultdict(lambda: 0)
            coeffs.update({x_c[x]: c for x, c in fnc_coeffs.items()})
            if isinstance(fnc, InEquality):
                coeffs[x_c[fnc.slack]] = 1 if fnc.op == Relations.ge else -1
            expr = template.subs([(c_, coeffs[c_]) for c_ in c])
            if isinstance(fnc, Equality):
                return Equality(expr, variables=x, vmap=x_v)
            elif isinstance(fnc, InEquality):
                eq = Equality(expr, variables=x, vmap=x_v)
                eq.slack = fnc.slack
                return eq
            elif isinstance(fnc, Function):
                return Function(expr, variables=x, vmap=x_v)
            else:
                return None

        obj = redefine(self.obj)
        constraints = [redefine(c) for c in self.constraints]
        return LLP(obj=obj, constraints=constraints, variables=x)

    def eval_constraints(self, x):
        return np.array([c.f0(x) for c in self.constraints],
                        dtype=np.float32)

    def feasible(self, x=None) -> bool:
        c = [c.relate(x) for c in self.constraints]
        return all(c) and all(x >= 0)

    @staticmethod
    def basic_solution(A=None, b=None, order=None, **kwargs):
        m, n = A.shape
        r = n - m
        assert r > 0

        stop = False
        try:
            if order is not None:
                if isinstance(order, Iterable):
                    permutations = iter([order])
            else:
                order = [i for i in range(n)]
                permutations = multiset_permutations(order)
            while not stop:
                order = next(permutations)
                A_ = A[:, order]
                B_ = A_[:, :m]
                try:
                    B_inv = np.linalg.inv(B_)
                    xB = np.matmul(B_inv, b)
                    stop = all(xB >= 0)
                except LinAlgError:
                    """
                    If there is no error, it means that calculation
                    of xB was succesful, which is only possible if the
                    current permutation defines a positive definite submatrix.
                    Note that this is cheaper than checking the eigenvalues,
                    since it only requires the eigenvalues to be all positive,
                    and does not involve calculating their actual values.
                    """
                    pass
        except StopIteration:
            """
            There is no permutation of columns that would produce a regular
            mxm submatrix
                -> there is no feasible basic solution
                    -> there is no feasible solution
            """
            pass
        finally:
            if stop:
                N_ = A_[:, m:]
                xN = np.zeros(r, dtype=np.float32)
                return B_, B_inv, N_, xB, xN, order
            else:
                return None

    @staticmethod
    def solve_standard_form(A, b, c, order=None, tol=1e-10):
        m, n = A.shape
        r = n - m
        assert r > 0
        basic = LLP.basic_solution(A, b, order=order)
        if basic:
            B, B_inv, N, xB, xN, order = basic
            c_ = c[order]
            cB = c_[:m]
            cN = c_[m:]
        else:
            return None

        def current_result(message: str = None, fail=False):
            if message:
                print(message)
            if not fail:
                return np.concatenate((xB, xN))[np.argsort(order)]
            return None

        def unit_basis_vector(length, index=0, value=1.0):
            return value * np.bincount([index], None, length)

        degenerate = False
        while True:
            if degenerate:
                # The objective could be decreased, but only on the expense
                # of violating positivity of the standard variables.
                # Hence, the solution is degenerate.
                return current_result('degenerate solution')

            # calculate reduced costs
            W = np.matmul(B_inv, N)
            reduced_costs = cN - np.matmul(cB, W)
            nEntering = np.count_nonzero(reduced_costs < 0)
            if nEntering == 0:
                # The objective can not be further reduced.
                # There was only one basic solution, which is
                # a unique optimizer.
                d = np.count_nonzero(reduced_costs > 0)
                if d < len(reduced_costs):
                    return current_result('multiple solutions')
                else:
                    return current_result('unique solution')
            # If we reach this line, reduction of the objective is possible,
            # although maybe indefinitely. If the objective can be decreased,
            # but only on the expense of violating feasibility, the
            # solution is degenerate.
            degenerate = True

            # Candidates for entering index are the indices of the negative
            # components of the vector of reduced costs.
            i_entering = np.argsort(reduced_costs)[:nEntering]
            for i_enter in i_entering:
                # basis vector for the current entering variable
                b_enter = unit_basis_vector(r, i_enter, 1.0)

                # w = vector of decrements of the current solution xB
                # Only positive values are a threat to feasibility, and we
                # need to tell which of the components of xB vanishes first,
                # which, since all components of xB are posotive,
                # has to do with the positive components only.
                w_enter = np.matmul(W, b_enter)
                i_leaving = np.argwhere(w_enter > 0)
                if len(i_leaving) == 0:
                    # step size could be indefinitely increased in this
                    # direction without violating feasibility, there is
                    # no solution to the problem
                    return current_result('no solution', fail=True)
                vanishing_ratios = xB[i_leaving]/w_enter[i_leaving]
                # the variable that vanishes first is the one with the smallest
                # vanishing ratio
                i_leave = i_leaving.flatten()[np.argmin(vanishing_ratios)]

                # step size in the direction of current basis vector
                t = xB[i_leave]/w_enter[i_leave]

                # update_solution
                if abs(t) <= tol:
                    # Smallest vanishing ratio is zero, any step would
                    # result in an infeasible situation.
                    # -> go for the next entering variable
                    continue
                xB -= t*w_enter
                xN = t*b_enter

                # reorder
                order[m + i_enter], order[i_leave] = \
                    order[i_leave], order[m + i_enter]
                B[:, i_leave], N[:, i_enter] = \
                    N[:, i_enter], copy(B[:, i_leave])
                B_inv = np.linalg.inv(B)
                cB[i_leave], cN[i_enter] = cN[i_enter], cB[i_leave]
                xB[i_leave], xN[i_enter] = xN[i_enter], xB[i_leave]

                # break loop at the first meaningful (t != 0) decrease and
                # force recalculation of the vector of reduced costs
                degenerate = False
                break

    def solve(self, order=None, as_dict=False):
        P = self.as_standard()

        # calculate A, b and c
        x = P.variables
        n = len(x)
        zeros = np.zeros((n,), dtype=np.float32)
        b = - P.eval_constraints(zeros)
        A = []
        for c in P.constraints:
            coeffs = c.linear_coefficients(normalize=True)
            A.append(np.array([coeffs[x_] for x_ in x], dtype=np.float32))
        A = np.vstack(A)
        coeffs = P.obj.linear_coefficients(normalize=True)
        c = np.array([coeffs[x_] for x_ in x], dtype=np.float32)

        # calculate solution
        x_ = LLP.solve_standard_form(A, b, c, order=order)
        if x_ is not None:
            self.vmanager.substitute({v: val
                                     for v, val in zip(P.variables, x_)})
            if as_dict:
                return self.vmanager.vmap
            else:
                return np.array(list(self.vmanager.vmap.values()),
                                dtype=np.float32)
        else:
            return None


if __name__ == '__main__':

    """
    Example for unique solution
    (0, 6, 0, 4) --> 10
    The following order automatically creates
    a feasble solution : [0, 2, 3, 1]
    """
    variables = ['x1', 'x2', 'x3', 'x4']
    x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    obj1 = Function(3*x1 + 9*x3 + x2 + x4, variables=syms)
    eq11 = Equality(x1 + 2*x3 + x4 - 4, variables=syms)
    eq12 = Equality(x2 + x3 - x4 - 2, variables=syms)
    P1 = LLP(cost=obj1, constraints=[eq11, eq12], variables=syms)
    x_1 = P1.solve(order=[0, 2, 3, 1])
    print(x_1)

    """
    Example for degenerate solution.
    (0, 2, 0, 0)
    """
    variables = ['x1', 'x2', 'x3', 'x4']
    x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    obj2 = Function(3*x1 + x2 + 9*x3 + x4, variables=syms)
    eq21 = Equality(x1 + 2*x3 + x4, variables=syms)
    eq22 = Equality(x2 + x3 - x4 - 2, variables=syms)
    P2 = LLP(cost=obj2, constraints=[eq21, eq22], variables=syms)
    x_2 = P2.solve()
    print(x_2)
    
    """
    Example for no solution.
    """
    variables = ['x1', 'x2', 'x3', 'x4']
    x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    obj3 = Function(-3*x1 + x2 + 9*x3 + x4, variables=syms)
    eq31 = Equality(x1 - 2*x3 - x4 + 2, variables=syms)
    eq32 = Equality(x2 + x3 - x4 - 2, variables=syms)
    P3 = LLP(cost=obj3, constraints=[eq31, eq32], variables=syms)
    x_3 = P3.solve()
    print(x_3)
    
    """
    Example for multiple solutions.
    (0, 1, 1, 0), (0, 4, 0, 2)
    """
    variables = ['x1', 'x2', 'x3', 'x4']
    x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    obj4 = Function(3*x1 + 2*x2 + 8*x3 + x4, variables=syms)
    eq41 = Equality(x1 - 2*x3 - x4 + 2, variables=syms)
    eq42 = Equality(x2 + x3 - x4 - 2, variables=syms)
    P4 = LLP(cost=obj4, constraints=[eq41, eq42], variables=syms)
    x_4 = P4.solve()
    print(x_4)


# -*- coding: utf-8 -*-
import numpy as np
from numba import jit


@jit(nopython=True, nogil=True)
def max_frontwidth_bulk(topo: np.ndarray, path: np.ndarray, nDOF: int = 1):
    nV = np.max(np.abs(topo))
    nE, nN = topo.shape
    ndest = np.zeros(nV, dtype=np.int8)
    maxwidth = 0
    width = 0
    for i in range(nE):
        iE = path[i]
        for j in range(nN):
            if ndest[abs(topo[iE, j]) - 1] == 0:
                width += nDOF
                ndest[abs(topo[iE, j]) - 1] = 1
        maxwidth = max(maxwidth, width)
        for j in range(nN):
            if topo[iE, j] < 0:
                width -= nDOF
    return maxwidth


@jit(nopython=True, nogil=True)
def max_frontwidth_flattened(topof):
    nV = np.max(np.abs(topof))
    ndest = np.zeros(nV, dtype=np.int8)
    maxwidth = 0
    width = 0
    for i in range(len(topof)):
        if ndest[abs(topof[i]) - 1] == 0:
            width += 1
            ndest[abs(topof[i]) - 1] = 1
        maxwidth = max(maxwidth, width)
        if topof[i] < 0:
            width -= 1
    return maxwidth


@jit(nopython=True, nogil=True)
def flatind(row: int, col: int, ncols: int, base: int = 0):
    return base + ncols * row + col


@jit(nopython=True, nogil=True)
def flatsize_sym(n: int):
    return np.int32((n * n - n) / 2 + n + 0.1)


@jit(nopython=True, nogil=True)
def flatind_sym(row: int, col: int, base: int = 0):
    return flatsize_sym(max(row, col) - base) + min(row, col)


@jit(nopython=True, nogil=True)
def flatind_bulk(rows: np.ndarray, cols: np.ndarray, base: int = 0):
    ncols = np.max(cols) - base
    ncase, ndata = cols.shape
    res = np.zeros((ncase, ndata), 0, dtype=np.int32)
    for i in range(ncase):
        for j in range(ndata):
            res[i, j] = flatind(rows[i, j], cols[i, j], ncols, base)
    return res


@jit(nopython=True, nogil=True)
def flatind_flattened(rows: np.ndarray, cols: np.ndarray, base: int = 0):
    ncols = np.max(cols) - base
    ndata = len(cols)
    res = np.zeros(ndata, 0, dtype=np.int32)
    for i in range(ndata):
        res[i] = flatind(rows[i], cols[i], ncols, base)
    return res
